/*
+-----------------------------------------------------------------------------+
| @BEGIN_COPYRIGHT                                                            |
|                                                                             |
| Copyright (C) 2001-2003, The Regents of the University of California        |
| All Rights Reserved.                                                        |
|                                                                             |
| Visualization and Interactive Systems Group                                 |
| University of California, Irvine                                            |
| Irvine, CA 92697                                                            |
|                                                                             |
| Commercial use of this program without express permission of the            |
| University of California, Irvine, is strictly prohibited. Information about |
| usage and redistribution, and a disclaimer of all warrenties are            |
| available in the Copyright file provided with this code.                    |
|                                                                             |
| Developed By:                                                               |
|    Gautam Chaudhary    {gchaudha@uci.edu}                                   |
|    Tara C. Hutchinson  {thutchin@uci.edu}                                   |
|    Falko Kuester       {fkuester@uci.edu}                                   |
|    Thomas Wishgoll     {twischgo@uci.edu}                                   |
|                                                                             |
| @END_COPYRIGHT                                                              |
+-----------------------------------------------------------------------------+
|                                               
| Component  : BcamDib
| Filename   : BcamDib.h
| Sourcefile : BcamDib.cpp
| Language   : C++
|                                               
+----------------------------- Revision Information---------------------------+
| $Author: wischgol $
| $Revision: 1.3 $
| $Date: 2003/09/25 15:29:40 $
| $RCSfile: BcamDib.h,v $
+-----------------------------------------------------------------------------+
*/


/*! \file	BcamDib.h 
    \brief	Interface and imeplementation of the CDib and CDibPtr classes

	Company:  Basler Vision Technologies
	Section:  Vision Components
	Project:  1394 Driver
	$Author: wischgol $
	$Revision: 1.3 $
	$Date: 2003/09/25 15:29:40 $
*/

#pragma once

//#define IPL_SUPPORT

//#include <atlgdi.h>
#include <bcam.h>
#include <assert.h>
#include "dibsectn.h"

/*! \def IPL_SUPPORT
	If this symbol is defined CDib supports IPL's (= Intel Performance Library)
	native image format IplImage.
*/
#ifdef IPL_SUPPORT
	#include <ipl.h>
#endif

using namespace Bcam;

class CDibPtr;

//! Encapsulates a Windows DIB section and implements some missing features
/*! The Windows GDI Bitmap object has two drawbacks:
	- It does not handle the image's orientation (bottom up or top down) properly
	- It does not know about monochrome palettes
	
	This class overcomes these drawbacks. It encapsulates a Windows GDI Bitmap handle
	and adds the missing functionality.

	The life time of CDib objects is controlled by a reference counter implementing the same semantics 
	as COM objects.

	This class is intended to be used only via the corresponding smart pointer #CDibPtr.
	To ensure this the constructor, the creator functions and the functions to 
	increase and decrease the reference counter are accessible through CDibPtr only.

*/
class CDib 
{
	friend CDibPtr;

// Public enums
public:
	//! Orientation of the image
	enum Orientation_t
	{
		BottomUp, //!< The bottom line of the image is located at lower adress than the top line
		TopDown   //!< The top line of the image is located at lower adress than the bottom line
	};

	//! The way the colors in the palette are choosen
	enum PaletteType_t
	{
		Color,		//!< The resulting image is a color image
		Monochrome	//!< The resulting image is a monochrome image
	};

// Protected creation functions anly to be accessed by CDibPtr
protected:
	//! Default constructor
	CDib() :
		m_hBitmap(NULL),
		m_Orientation(BottomUp),
		m_PaletteType(Color),
		m_RefCount(0)
	{
#ifdef IPL_SUPPORT
		m_pIplImage = NULL;
#endif
	}

	//! Destroy all BMP related resources 
	void Destroy() 
	{
		// Destroy the referenced DIB section
		if(m_hBitmap)
		{
			if(!::DeleteObject(m_hBitmap))
				throw  BcamException(::GetLastError(), "CDib::Destroy");
			m_hBitmap = NULL;
		}

#ifdef IPL_SUPPORT

		// Destroy the IPL image header
		if(m_pIplImage)
		{
			iplDeallocate( m_pIplImage, IPL_IMAGE_HEADER );
			m_pIplImage = NULL;
		}

#endif // #ifdef IPL_SUPPORT

	}

	//! Increments the reference counter of this object
	long AddRef()
	{
		return ::InterlockedIncrement(&m_RefCount);
	}

	//! Decrements the reference counter of this object. Destroys the object if the reference counter becomes 0.
	void Release()
	{
		::InterlockedDecrement(&m_RefCount);
		if(0 == m_RefCount)
			delete this;
	}

	//! Create a CDib object
	static CDib* Create( CSize Size, unsigned int BitsPerPixel, Orientation_t Orientation, PaletteType_t PaletteType ) 
	{
		CDib *pDib = new CDib();

		pDib->m_Orientation = Orientation;
		pDib->m_PaletteType = PaletteType;

		pDib->m_hBitmap = DSCreateDIBSection( Size.cx, Orientation == BottomUp ? Size.cy : -Size.cy, BitsPerPixel, PaletteType == Monochrome );
		if(!pDib->m_hBitmap)
			throw  BcamException(E_FAIL, "CDib::Create");

		pDib->AddRef();

		pDib->Refresh();

		return	pDib;
	}

public:
	//!  Loads the bitmap from a BMP file
	static CDib* LoadBMP( CString FileName )
	{
		CDib *pDib = new CDib();

		if(!DSLoadDIBSectionFromBMPFile( (LPCTSTR) FileName, &pDib->m_hBitmap, NULL ))
			throw  BcamException(E_FAIL, "CDib::LoadBMP");

		// BMP files are always bottom up
		pDib->m_Orientation = BottomUp;

		// Check if the image is Mono8
		pDib->m_PaletteType = Color;
		LPBITMAPINFO pBitmapInfo = DSGetBITMAPINFOForDIBSection( pDib->m_hBitmap );
		if(pBitmapInfo->bmiHeader.biBitCount == 8)
		{
			for(int i=0; i<256; ++i)
			{
				if(    pBitmapInfo->bmiColors[i].rgbRed != i
					|| pBitmapInfo->bmiColors[i].rgbGreen != i
					|| pBitmapInfo->bmiColors[i].rgbBlue != i )
					break;
			}
			pDib->m_PaletteType = Monochrome;
		}

		free(pBitmapInfo);

		pDib->AddRef();

		pDib->Refresh();

		return pDib;
	}

	//! Virtual destructor
	virtual ~CDib() 
	{
		Destroy();
	}

	void SetOrientation (Orientation_t orientation)
	{
		m_Orientation = orientation;
	}

	//!  Stores the bitmap to a BMP file
	void StoreBMP( CString FileName )
	{
		if( m_Orientation == TopDown )
		{
			// BMP file have to be stores bottom up so we have to topple it
			HBITMAP hBitmap = GetToppledCopy();
			DSStoreDIBSectionInBMPFile( (LPCTSTR) FileName, hBitmap);
			::DeleteObject(hBitmap);
		}
		else
		{
			DSStoreDIBSectionInBMPFile( (LPCTSTR) FileName, m_hBitmap);
		}
	}

	//! Gets a HBITMAP handle (no ownership is granted!)
	HBITMAP GetBitmapHandle() 
	{ 
		return m_hBitmap; 
	}

	//!  Returns a pointer to the pixel data 
	void* GetPixels()
	{
		BITMAP Bitmap;
		if(!::GetObject(m_hBitmap, sizeof(BITMAP), &Bitmap))
			throw  BcamException(::GetLastError(), "CDib::GetPixels");

		return Bitmap.bmBits;
	}

	//!  Returns the bitmap's size 1
	void GetSize(CSize *pSize)
	{
		assert(pSize);
		if(!m_hBitmap)
		{
			pSize->cx = 0;
			pSize->cy = 0;
		}
		else
		{
			BITMAP Bitmap;
			if(!::GetObject(m_hBitmap, sizeof(BITMAP), &Bitmap))
				throw  BcamException(::GetLastError(), "CDib::GetSize");

			pSize->cx = Bitmap.bmWidth;
			pSize->cy = Bitmap.bmHeight;
		}
	}

	//!  Returns the bitmap's size 2
	CSize GetSize()
	{
		CSize Size;
		GetSize(&Size);
		return Size;
	}

	//!  Returns the total size of a line in bytes
	unsigned long GetWidthBytes()
	{
		BITMAP Bitmap;
		if(!::GetObject(m_hBitmap, sizeof(BITMAP), &Bitmap))
			throw  BcamException(::GetLastError(), "CDib::GetSize");

		return Bitmap.bmWidthBytes;
	}

	//!  Returns the total size of the image in bytes
	unsigned long GetTotalPixelBytes()
	{
		BITMAP Bitmap;
		if(!::GetObject(m_hBitmap, sizeof(BITMAP), &Bitmap))
			throw  BcamException(::GetLastError(), "CDib::GetSize");

		return Bitmap.bmWidthBytes * Bitmap.bmHeight;
	}

	//!  Returns the bitmap's bits per pixel
	unsigned short GetBitsPerPixel()
	{
		BITMAP Bitmap;
		if(!::GetObject(m_hBitmap, sizeof(BITMAP), &Bitmap))
			throw  BcamException(::GetLastError(), "CDib::GetBitsPerPixel");

		return Bitmap.bmBitsPixel;
	}

	//! Returns the bitmaps orientation (BottomUp or TopDown)
	Orientation_t GetOrientation()
	{
		return m_Orientation;
	}

	//! Returns the bitmaps palette type (e.g., color or monochrome )
	PaletteType_t GetPaletteType()
	{
		return m_PaletteType;
	}

	//! Returns the bitmaps BITMAPINFORHEADER 
	void GetBitmapInfoHeader(BITMAPINFOHEADER *pBitmapInfoHeader)
	{
		DIBSECTION DibSection;
		if(!::GetObject(m_hBitmap, sizeof(DIBSECTION), &DibSection))
			throw  BcamException(::GetLastError(), "CDib::GetBitmapInfoHeader");

		::CopyMemory(pBitmapInfoHeader, &(DibSection.dsBmih), sizeof(BITMAPINFOHEADER));
	}

	//! Returns a bitmaps BITMAPINFO which has to be freed by the user
	void GetBitmapInfo(BITMAPINFO **ppBitmapInfo, unsigned long *pBitmapInfoBytes)
	{
		
		*ppBitmapInfo = DSGetBITMAPINFOForDIBSection( m_hBitmap );
		*pBitmapInfoBytes = sizeof(BITMAPINFOHEADER) + DSColorTableSize( *ppBitmapInfo );
	}

	//! Returns the bitmaps BITMAP 
	void GetBitmap(BITMAP *pBitmap)
	{
		if(!::GetObject(m_hBitmap, sizeof(BITMAP), pBitmap))
			throw  BcamException(::GetLastError(), "CDib::GetBitmap");
	}

	//!  Topples the bitmap
	void Topple()
	{
		if (m_hBitmap == NULL)
			throw  BcamException(E_FAIL, "CDib::Topple - No bitmap present");

		HBITMAP hBitmap = GetToppledCopy();
		Destroy();

		m_hBitmap = hBitmap;

		m_Orientation = m_Orientation == BottomUp ? TopDown : BottomUp;

		Refresh();

	}

	//! Clones the Bitmap 
	CDib* Clone(
		bool Topple = false  //!< If true the clone will be toppled with resprect to the original bitmap
		)
	{
		if (m_hBitmap == NULL)
			throw  BcamException(E_FAIL, "CDib::Clone - No bitmap present");

		HBITMAP pBitmap = NULL;
		if(Topple)
		{
			pBitmap = GetToppledCopy();
		}
		else
		{
			pBitmap = GetCopy();
		}
		
		CDib *pDib = new CDib();
		pDib->m_hBitmap = pBitmap;
		pDib->m_Orientation = Topple ? (m_Orientation == BottomUp ? TopDown : BottomUp) : m_Orientation;
		pDib->m_PaletteType = m_PaletteType;
		pDib->AddRef();
		
		pDib->Refresh();

		return pDib;
	}

	//! Copies the image to the clipboard
	void CopyToClipboard()
	{
		if ( GetOrientation() == CDib::TopDown )
		{
			CDib* pDib = Clone(true);
			if (pDib == NULL)
				throw  BcamException(E_FAIL, "CDib::CopyToClipboard");
			pDib->CopyToClipboard();
			delete pDib;
		}
		else
		{
			HGLOBAL hGlobal = NULL;
			try
			{
				if (m_hBitmap == NULL)
					throw  BcamException(E_FAIL, "CDib::CopyToClipboard - No bitmap present");

				// Get BITMAPINFO Header
				LPBITMAPINFO pBitmapInfo;
				pBitmapInfo = DSGetBITMAPINFOForDIBSection( m_hBitmap );
				if(!pBitmapInfo)
					throw  BcamException(::GetLastError(), "CDib::CopyToClipboard");

				// Create a suitable global memory buffer and lock it 
				DWORD TotalDIBBytes = DSTotalBytesSize( pBitmapInfo );
				hGlobal = GlobalAlloc(GHND | GMEM_SHARE, TotalDIBBytes );
				if(!hGlobal)
					throw  BcamException(::GetLastError(), "CDib::CopyToClipboard");

				// Lock the globaal Buffer
				char* pBuffer = (char*) GlobalLock(hGlobal);
				if ( pBuffer == NULL )
					throw  BcamException(::GetLastError(), "CDib::CopyToClipboard");

				// Copy the information header and the color table to the global Buffer
				DWORD dwHeaderSize = sizeof(BITMAPINFOHEADER) + DSColorTableSize( pBitmapInfo );
				CopyMemory(pBuffer, &pBitmapInfo->bmiHeader, dwHeaderSize);
				
				// Copy the image data to the global buffer
				LPBYTE pBits = DSGetPointerToDIBSectionImageBits(  m_hBitmap  );
				DWORD dwSize = DSImageBitsSize( pBitmapInfo );
				CopyMemory(((PBYTE) pBuffer) + dwHeaderSize, pBits, dwSize);

				// Clean up
				free(pBitmapInfo);
				GlobalUnlock(hGlobal);
					
				// set the buffer to the clipboard
				if(!::OpenClipboard(NULL))
					throw  BcamException(::GetLastError(), "CDib::CopyToClipboard");

				if(!::EmptyClipboard())
					throw  BcamException(::GetLastError(), "CDib::CopyToClipboard");

				if(!::SetClipboardData(CF_DIB, hGlobal))
					throw  BcamException(::GetLastError(), "CDib::CopyToClipboard");

				if(!::CloseClipboard())
					throw  BcamException(::GetLastError(), "CDib::CopyToClipboard");
			}
			catch(BcamException e)
			{
				if(hGlobal)
					::GlobalFree(hGlobal);
				throw e;
			}
		}
	}

#ifdef IPL_SUPPORT
	//! Copies the image to the clipboard
	IplImage* GetIplImage()
	{
		return m_pIplImage;
	}

#endif // #ifdef IPL_SUPPORT

// Helper functions
protected:
	//!  Returns a DIB section which is a copy of the referenced bitmap
	HBITMAP GetCopy()
	{
		if (m_hBitmap == NULL)
			throw  BcamException(E_FAIL, "CDib::GetCopy - No bitmap present");
		
		LPBITMAPINFO pBitmap = DSGetBITMAPINFOForDIBSection(m_hBitmap);
		ATLASSERT(pBitmap != NULL);
		pBitmap->bmiHeader.biHeight = m_Orientation == TopDown ? -pBitmap->bmiHeader.biHeight : pBitmap->bmiHeader.biHeight;

		LPBYTE pSrc = DSGetPointerToDIBSectionImageBits(m_hBitmap);
		ATLASSERT(pSrc != NULL);

		LPBYTE pDst = NULL;
		HBITMAP hBitmap = ::CreateDIBSection( NULL, pBitmap, DIB_RGB_COLORS, (void**)&pDst, NULL, 0 );
		free(pBitmap);

		if(hBitmap == NULL || pDst == NULL)
			throw  BcamException(::GetLastError(), "CDib::GetCopy");

		return hBitmap;
	}

	//!  Returns a DIB section which is a toppled copy of the referenced bitmap
	HBITMAP GetToppledCopy()
	{
		if (m_hBitmap == NULL)
			throw  BcamException(E_FAIL, "CDib::GetToppledCopy - No bitmap present");
		
		LPBITMAPINFO pBitmap = DSGetBITMAPINFOForDIBSection(m_hBitmap);
		ATLASSERT(pBitmap != NULL);

		LPBYTE pSrc = DSGetPointerToDIBSectionImageBits(m_hBitmap);
		ATLASSERT(pSrc != NULL);

		// Create a version of the image which is top down if the original image was bottom up 
		// and vice versa. 
		// BEWARE : For some strange reason the newly created image will have a biHeight > 0
		// even if we created it with biHeight < 0. Nevertheless the bitmap object somehow
		// remembers the fact that it is top-down because if it's drawn it appears correct
		pBitmap->bmiHeader.biHeight = -pBitmap->bmiHeader.biHeight;

		LPBYTE pDst = NULL;
		HBITMAP hBitmap = ::CreateDIBSection( NULL, pBitmap, DIB_RGB_COLORS, (void**)&pDst, NULL, 0 );
		if(hBitmap == NULL || pDst == NULL)
			throw  BcamException(::GetLastError(), "CDib::GetToppledCopy");

		DWORD rowlength = GetWidthBytes();
		
		for( int i = 0, m = abs(pBitmap->bmiHeader.biHeight) - 1;  i < abs(pBitmap->bmiHeader.biHeight); i++, m-- )
		{
			memcpy(pDst + m * rowlength, pSrc + i * rowlength, rowlength);
		}
		
		return hBitmap;
	}

	//!  This function is called whenever the bitmap is re-created
	void Refresh()
	{

		// TODO : Create a matching palette (but honestly : who is using
		// palettised graphic cards any more? :-)

#ifdef IPL_SUPPORT

	// Create a IPL header describing the BMP. No additional memory is allocated
	if( m_hBitmap )
	{

		BITMAP Bitmap;
		GetBitmap(&Bitmap);

		int nChannels = Bitmap.bmBitsPixel == 1 ? 1 : Bitmap.bmBitsPixel/8;
		int depth     = Bitmap.bmBitsPixel == 1 ? IPL_DEPTH_1U : IPL_DEPTH_8U;
		int origin	  = m_Orientation == BottomUp ? IPL_ORIGIN_BL : IPL_ORIGIN_TL;
    
		m_pIplImage = iplCreateImageHeader(
			nChannels,      // number of channels
			0,              // # of alpha channel, 0-absent
			depth,          // bit depth of channel
			"RGB",          // color model
			"BGR",          // channel sequence
			IPL_DATA_ORDER_PIXEL,// data ordering : by pixels or by channels
			origin,		    // bitmap origin : _TL = top left, _BL = bottom left
			IPL_ALIGN_DWORD,// scan lines alignment
			Bitmap.bmWidth,    // width in pixels
			Bitmap.bmHeight,   // height and 
			0,              // ROI, 0 - entire image  
			0,              // MASK ROI, 0 - entire image  
			0,              // image ID - reserved
			0);             // tile info.
    
		m_pIplImage->imageData = m_pIplImage->imageDataOrigin = (char*)(Bitmap.bmBits);
	}

#endif // #ifdef IPL_SUPPORT
	}


// Member Variables
protected:
	//! Handle of the referenced Bitmap
	HBITMAP m_hBitmap;

	//! Indicates the orientation of the image (topdown or bottomup)
	/*! In Windows a Bitmap object can be created topdown by setting Height < 0.
		For some strange reason the BITMAP structure returned from a topdown
		bitmap will have Height > 0 so the information has to be cached here.
	*/
	Orientation_t m_Orientation;

	//! Indicates if the bitmap is monochrome or color
	/*! In Windows a monochrome Bitmap is a 8 bit palletized image with a special
	    palette : RGB(i) = {i,i,i}. This information is cached.
	*/
	PaletteType_t m_PaletteType;

	//! Reference counter controlling the life of the CDib object 
	//! Same semantic as in COM objects
	long m_RefCount;

#ifdef IPL_SUPPORT
	//! Pointer to a header describing the bitmap as IPL image
	IplImage* m_pIplImage;

#endif 

};

//! Smart pointer to CDib objects
/*! This class implements the same semantics as a COM smart pointer.
	In addition it provides creator functions to create CDib objects.
*/
class CDibPtr
{
public:
	//! Default constructor
	CDibPtr()
	{
		m_pDib = NULL;
	}

	//! Copy constructor 1
	CDibPtr(CDibPtr* pDibPtr)
	{
		if(pDibPtr != NULL)
		{
			if ((m_pDib = pDibPtr->m_pDib) != NULL)
				m_pDib->AddRef();
		}
		else
			m_pDib = NULL;
	}

	//! Copy constructor 2
	CDibPtr(CDibPtr& DibPtr)
	{
		if ((m_pDib = DibPtr.m_pDib) != NULL)
			m_pDib->AddRef();
	}

	//! Creates a bismap object
	void Create( CSize Size, 
				 unsigned int BitsPerPixel, 
				 CDib::Orientation_t Orientation = CDib::BottomUp, 
				 CDib::PaletteType_t PaletteType = CDib::Color) 
	{
		Release();

		m_pDib = CDib::Create( Size, BitsPerPixel, Orientation, PaletteType );
	}

	//!  Loads a bitmap from a BMP file
	void LoadBMP( CString FileName )
	{
		Release();

		m_pDib = CDib::LoadBMP( FileName );
	}

	//! virtual Destructor
	virtual ~CDibPtr()
	{
		if(m_pDib)
			m_pDib->Release();
	}

	//! releases the referenced CDib object
	void Release()
	{
		CDib* pTemp = m_pDib;
		if (pTemp)
		{
			m_pDib = NULL;
			pTemp->Release();
		}
	}

	//! Gets access to the referenced CDib object
	CDib& operator*() const
	{
		if(!m_pDib)
			throw  BcamException(E_POINTER, "CDibPtr::operator*");

		return *m_pDib;
	}

	//! Gets access to the referenced CDib object
	CDib* operator->() const
	{
		if(!m_pDib)
			throw  BcamException(E_POINTER, "CDibPtr::operator->");

		return m_pDib;
	}

	//! Assigns a CDib pointer 
	CDibPtr& operator=(CDib* pDib)
	{
		Release();

		if(pDib)
		{
			m_pDib = pDib;
			if(m_pDib)
				m_pDib->AddRef();
		}
		else
			m_pDib = NULL;

		return *this;
	}

	//! Assigns a smart pointer 1
	CDibPtr& operator=(CDibPtr* pDibPtr)
	{
		Release();

		if(pDibPtr)
		{
			m_pDib = pDibPtr->m_pDib;
			if(m_pDib)
				m_pDib->AddRef();
		}
		else
			m_pDib = NULL;

		return *this;
	}

	//! Assigns a smart pointer 2
	CDibPtr& operator=(CDibPtr& DibPtr)
	{
		Release();

		m_pDib = DibPtr.m_pDib;
		if(m_pDib)
			m_pDib->AddRef();

		return *this;
	}

	//! True, if this smart pointer does not reference a CDib object
	bool IsNull() 
	{
		return (m_pDib == NULL);
	}

	//! True, if this smart pointer references a CDib object
	operator bool() 
	{
		return !IsNull();
	}

	//! True, if this smart pointer does not reference a CDib object
	bool operator!()
	{
		return IsNull();
	}

	//! Compares if two smar pointers are equal
	bool operator==(CDibPtr& DibPtr) const
	{
		return m_pDib == DibPtr.m_pDib;
	}
	
	//! Pointer to the referenced CDib object
	CDib* m_pDib;
};



